#!/usr/bin/env python
# encoding: utf-8
"""A service to sync a local file tree to jottacloud.

Copies and updates files in the cloud by comparing md5 hashes, like the official client.
Run it from crontab at an appropriate interval.

"""
# This file is part of jottacloudclient.
#
# jottacloudclient is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jottacloudclient is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jottacloudclient.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015 Håvard Gulldahl <havard@gulldahl.no>

#import included batteries
import os, re, os.path, sys, logging, argparse
import math, time

log = logging.getLogger(__name__)

#import pip modules
from clint.textui import progress, puts, colored

#import jottalib
from jottalib.JFS import JFS
from . import jottacloud, __version__


if sys.platform != "win32":
    # change progress indicators to something that looks nice
    #TODO: rather detect utf-8 support in the terminal
    #TODO: change this when https://github.com/kennethreitz/clint/pull/151 is merged
    progress.BAR_EMPTY_CHAR =u'○'
    progress.BAR_FILLED_CHAR=u'●'


def humanizeFileSize(size):
    size = abs(size)
    if (size==0):
        return "0B"
    units = ['B','KiB','MiB','GiB','TiB','PiB','EiB','ZiB','YiB']
    p = math.floor(math.log(size, 2)/10)
    return "%.3f%s" % (size/math.pow(1024,p),units[int(p)])

def filescanner(topdir, jottapath, jfs, errorfile, exclude=None, dry_run=False, prune_files=True, prune_folders=True ):

    errors = {}
    def saferun(cmd, *args):
        log.debug('running %s with args %s', cmd, args)
        try:
            return apply(cmd, args)
        except Exception as e:
            puts(colored.red('Ouch. Something\'s wrong with "%s":' % args[0]))
            log.exception('SAFERUN: Got exception when processing %s', args)
            errors.update( {args[0]:e} )
            return False

    _files = 0

    try:
        for dirpath, onlylocal, onlyremote, bothplaces, onlyremotefolders in jottacloud.compare(topdir, jottapath, jfs, exclude_patterns=exclude):
            puts(colored.green("Entering dir: %s" % dirpath))
            if len(onlylocal):
                _start = time.time()
                _uploadedbytes = 0
                for f in progress.bar(onlylocal, label="uploading %s new files: " % len(onlylocal)):
                    if os.path.islink(f.localpath):
                        log.debug("skipping symlink: %s", f)
                        continue
                    log.debug("uploading new file: %s", f)
                    if not dry_run:
                        if saferun(jottacloud.new, f.localpath, f.jottapath, jfs) is not False:
                            _uploadedbytes += os.path.getsize(f.localpath)
                            _files += 1
                _end = time.time()
                puts(colored.magenta("Network upload speed %s/sec" % ( humanizeFileSize( (_uploadedbytes / (_end-_start)) ) )))

            if prune_files and len(onlyremote):
                puts(colored.red("Deleting %s files from JottaCloud because they no longer exist locally " % len(onlyremote)))
                for f in progress.bar(onlyremote, label="deleting JottaCloud file: "):
                    log.debug("deleting cloud file that has disappeared locally: %s", f)
                    if not dry_run:
                        if saferun(jottacloud.delete, f.jottapath, jfs) is not False:
                            _files += 1
            if len(bothplaces):
                for f in progress.bar(bothplaces, label="comparing %s existing files: " % len(bothplaces)):
                    log.debug("checking whether file contents has changed: %s", f)
                    if not dry_run:
                        if saferun(jottacloud.replace_if_changed, f.localpath, f.jottapath, jfs) is not False:
                            _files += 1
            if prune_folders and len(onlyremotefolders):
                puts(colored.red("Deleting %s folders from JottaCloud because they no longer exist locally " % len(onlyremotefolders)))
                for f in onlyremotefolders:
                    if not dry_run:
                        if saferun(jottacloud.deleteDir, f.jottapath, jfs) is not False:
                            logging.debug("Deleted remote folder %s", f.jottapath)
    except KeyboardInterrupt:
        # Ctrl-c pressed, cleaning up
        pass
    if len(errors) == 0:
        puts('Finished syncing %s files to JottaCloud, no errors. yay!' % _files)
    else:
        puts(('Finished syncing %s files, ' % _files )+
             colored.red('with %s errors (read %s for details)' % (len(errors), errorfile, )))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''Mount your JottaCloud files locally and use it with your normal file tools'''
#
# This file is part of jottafs.
#
# jottafs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jottafs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jottafs.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2011,2013-2015 Håvard Gulldahl <havard@gulldahl.no>

# metadata

__author__ = 'havard@gulldahl.no'

# importing stdlib
import sys, os, pwd, stat, errno
import urllib, logging, datetime, argparse
import time
import itertools
try:
    from cStringIO import StringIO # py2
except ImportError:
    from io import StringIO # py3

logging.captureWarnings(True)
log = logging.getLogger(__name__)

# import jotta
from jottalib import JFS, __version__
from jottalib.contrib.mwt import Memoize

# import dependenceis (get them with pip!)
try:
    from fuse import FUSE, Operations, LoggingMixIn, FuseOSError # this is 'pip install fusepy'
except ImportError:
    print "JottaFuse won't work without fusepy! Please run `pip install fusepy`."
    raise

class JottaFuseError(FuseOSError):
    pass


ESUCCESS=0

BLACKLISTED_FILENAMES = ('.hidden', '._', '._.', '.DS_Store',
                         '.Trash', '.Spotlight-', '.hotfiles-btree',
                         'lost+found', 'Backups.backupdb', 'mach_kernel')

def is_blacklisted(path):
    _basename = os.path.basename(path)
    for bf in BLACKLISTED_FILENAMES:
        if _basename.startswith(bf):
            return True
    return False


class JottaFuse(LoggingMixIn, Operations):
    '''
    A simple filesystem for JottaCloud.
    '''


    def __init__(self, auth, path='.'):
        self.client = JFS.JFS(auth)
        self.__newfiles = {} # a dict of stringio objects
        self.__newfolders = []
        self.ino = 0

    #
    # setup and teardown
    #
    def init(self, rootpath):
        # Called on filesystem initialization. (Path is always /)
        # Use it instead of __init__ if you start threads on initialization.
        #TODO: Set up threaded work queue
        pass

    def destroy(self, path):
        #TODO: do proper teardown
        pass


    #
    # helpers
    #

    def _getpath(self, path):
        "A wrapper of JFS.getObject(), with some tweaks that make sense in a file system."
        if is_blacklisted(path):
            raise JottaFuseError('Blacklisted file, refusing to retrieve it')

        return self.client.getObject(path)

    def _dirty(self, path):
        'Remove path from cache'
        return Memoize().yank_path(path)

    #
    # some methods are expected to always work on a rw filesystem, so let's make them work
    #

    def _success(self, *args):
        '''shortcut to always return success (0) to masquerade as a proper filesystem'''
        return 0


    chmod = _success
    chown = _success
    utimens = _success
    setxattr = _success


    #
    # fuse syscall implementations
    #


    def create(self, path, mode, fi=None):
        if is_blacklisted(path):
            raise JottaFuseError('Blacklisted file')
        self.__newfiles[path] = StringIO()
        self.ino += 1
        return self.ino

    @Memoize(timeout=60) # remember every result for 60 seconds
    def getattr(self, path, fh=None):
        if is_blacklisted(path):
            raise OSError(errno.ENOENT)
        pw = pwd.getpwuid( os.getuid() )
        if path in self.__newfolders: # folder was just created, not synced yet
            return {
                'st_atime': time.time(),
                'st_gid': pw.pw_gid,
                'st_mode': stat.S_IFDIR | 0755,
                'st_mtime': time.time(),
                'st_size': 0,
                'st_uid': pw.pw_uid,
                }
        elif path in self.__newfiles: # file was just created, not synced yet
            return {
                'st_atime': time.time(),
                'st_gid': pw.pw_gid,
                'st_mode': stat.S_IFREG | 0644,
                'st_mtime': time.time(),
                'st_size': 0,
                'st_uid': pw.pw_uid,
                }
        try:
            f = self._getpath(path)
        except JFS.JFSError:
            raise OSError(errno.ENOENT, '') # file not found
        if isinstance(f, (JFS.JFSFile, JFS.JFSFolder, JFS.JFSIncompleteFile)) and f.is_deleted():
            raise OSError(errno.ENOENT)

        if isinstance(f, (JFS.JFSIncompleteFile, JFS.JFSFile)):
            _mode = stat.S_IFREG | 0644
        elif isinstance(f, JFS.JFSFolder):
            _mode = stat.S_IFDIR | 0755
        elif isinstance(f, (JFS.JFSMountPoint, JFS.JFSDevice) ):
            _mode = stat.S_IFDIR | 0555 # these are special jottacloud dirs, make them read only
        else:
            if not f.tag in ('user', ):
                log.warning('Unknown jfs object: %s <-> "%s"' % (type(f), f.tag) )
            _mode = stat.S_IFDIR | 0555
        return {
                'st_atime': time.mktime(f.modified.timetuple()) if isinstance(f, JFS.JFSFile) else time.time(),
                'st_gid': pw.pw_gid,
                'st_mode': _mode,
                'st_mtime': time.mktime(f.modified.timetuple()) if isinstance(f, JFS.JFSFile) else time.time(),
                'st_size': f.size if isinstance(f, JFS.JFSFile) else 0,
                'st_uid': pw.pw_uid,
                }

    def mkdir(self, path, mode):
        parentfolder = os.path.dirname(path)
        newfolder = os.path.basename(path)
        try:
            f = self._getpath(parentfolder)
        except JFS.JFSError:
            raise OSError(errno.ENOENT, '')
        if not isinstance(f, JFS.JFSFolder):
            raise OSError(errno.EACCES) # can only create stuff in folders
        if isinstance(f, (JFS.JFSFile, JFS.JFSFolder)) and f.is_deleted():
            raise OSError(errno.ENOENT)
        r = f.mkdir(newfolder)
        self.__newfolders.append(path)
        self._dirty(path)
        return ESUCCESS

    def open(self, path, flags):
        if flags & os.O_WRONLY:
            if not self.__newfiles.has_key(path):
                self.__newfiles[path] = StringIO()
            self.ino += 1
            return self.ino
        return super(JottaFuse, self).open(path, flags)

    def read(self, path, size, offset, fh):
        if path in self.__newfiles.keys(): # file was just created, not synced yet
            data = StringIO(self.__newfiles[path].getvalue())
            data.seek(offset, 0)
            buf = data.read(size)
            data.close()
            return buf
        else:
            try:
                f = self._getpath(path)
            except JFS.JFSError:
                raise OSError(errno.ENOENT, '')
            if isinstance(f, (JFS.JFSFile, JFS.JFSFolder)) and f.is_deleted():
                raise OSError(errno.ENOENT)
            # gnu tools may happily ask for content beyond file size
            # but jottacloud doesn't like that
            # so we make sure we stay within file size (f.size)
            end = min(offset+size, f.size)
            log.debug("f.readpartial(%s, %s) on file of size %s" % (offset, end, f.size))
            return f.readpartial(offset, end)

    def readdir(self, path, fh):
        yield '.'
        yield '..'
        if path == '/':
            for d in self.client.devices:
                yield d.name
        else:
            p = self._getpath(path)
            if isinstance(p, JFS.JFSDevice):
                for name in p.mountPoints.keys():
                    yield name
            else:
                for el in itertools.chain(p.folders(), p.files()):
                    if not el.is_deleted():
                        yield el.name

    def release(self, path, fh):
        "Run after a read or write operation has finished. This is where we upload on writes"
        #print "release! inpath:", path in self.__newfiles.keys()
        # if the path exists in self.__newfiles.keys(), we have a new version to upload
        try:
            f = self.__newfiles[path] # make a local shortcut to Stringio object
            f.seek(0, os.SEEK_END)
            if f.tell() > 0: # file has length
                self.client.up(path, f) # upload to jottacloud
            del self.__newfiles[path]
            del f
            self._dirty(path)
        except KeyError:
            pass
        return ESUCCESS

    def rename(self, old, new):
        if old == new: return
        try:
            f = self._getpath(old)
        except JFS.JFSError:
            raise OSError(errno.ENOENT, '')
        f.rename(new)
        self._dirty(old)
        self._dirty(new)
        return ESUCCESS

    @Memoize(timeout=60)
    def statfs(self, path):
        "Return a statvfs(3) structure, for stat and df and friends"
        # from fuse.py source code:
        #
        # class c_statvfs(Structure):
        # _fields_ = [
        # ('f_bsize', c_ulong), # preferred size of file blocks, in bytes
        # ('f_frsize', c_ulong), # fundamental size of file blcoks, in bytes
        # ('f_blocks', c_fsblkcnt_t), # total number of blocks in the filesystem
        # ('f_bfree', c_fsblkcnt_t), # number of free blocks
        # ('f_bavail', c_fsblkcnt_t), # free blocks avail to non-superuser
        # ('f_files', c_fsfilcnt_t), # total file nodes in file system
        # ('f_ffree', c_fsfilcnt_t), # free file nodes in fs
        # ('f_favail', c_fsfilcnt_t)] #
        #
        # On Mac OS X f_bsize and f_frsize must be a power of 2
        # (minimum 512).

        _blocksize = 512
        _usage = self.client.usage
        _fs_size = self.client.capacity
        if _fs_size == -1: # unlimited
            # Since backend is supposed to be unlimited,
            # always return a half-full filesystem, but at least 1 TB)
            _fs_size = max(2 * _usage, 1024 ** 4)
        _bfree = ( _fs_size - _usage ) // _blocksize
        return {
            'f_bsize': _blocksize,
            'f_frsize': _blocksize,
            'f_blocks': _fs_size // _blocksize,
            'f_bfree': _bfree,
            'f_bavail': _bfree,
            # 'f_files': c_fsfilcnt_t,
            # 'f_ffree': c_fsfilcnt_t,
            # 'f_favail': c_fsfilcnt_t

        }

    def symlink(self, linkname, existing_file):
        """Called to create a symlink `target -> source` (e.g. ln -s existing_file linkname). In jottafuse, we upload the _contents_ of source.

        This is a handy shortcut for streaming uploads directly from disk, without reading the file
        into memory first"""
        log.info("***SYMLINK* %s (link) -> %s (existing)", linkname, existing_file)
        sourcepath = os.path.abspath(existing_file)
        if not os.path.exists(sourcepath): # broken symlink
            raise OSError(errno.ENOENT, '')
        try:
            with open(sourcepath) as sourcefile:
                self.client.up(linkname, sourcefile)
                return ESUCCESS
        except Exception as e:
            log.exception(e)

        raise OSError(errno.ENOENT, '')

    def truncate(self, path, length, fh=None):
        "Download existing path, truncate and reupload"
        try:
            f = self._getpath(path)
        except JFS.JFSError:
            raise OSError(errno.ENOENT, '')
        if isinstance(f, (JFS.JFSFile, JFS.JFSFolder)) and f.is_deleted():
            raise OSError(errno.ENOENT)
        data = StringIO(f.read())
        data.truncate(length)
        try:
            self.client.up(path, data) # replace file contents
            self._dirty(path)
            return ESUCCESS
        except:
            raise OSError(errno.ENOENT, '')

    def rmdir(self, path):
        if path in self.__newfolders: # folder was just created, not synced yet
            self.__newfolders.remove(path)
            self._dirty(path)
            return
        try:
            f = self._getpath(path)
        except JFS.JFSError:
            raise OSError(errno.ENOENT, '')
        r = f.delete()
        self._dirty(path)
        return ESUCCESS

    def unlink(self, path):
        if path in self.__newfiles.keys(): # file was just created, not synced yet
            del self.__newfiles[path]
            self._dirty(path)
            return
        try:
            f = self._getpath(path)
        except JFS.JFSError:
            raise OSError(errno.ENOENT, '')
        r = f.delete()
        self._dirty(path)
        return ESUCCESS

    def write(self, path, data, offset, fh=None):
        if is_blacklisted(path):
            raise JottaFuseError('Blacklisted file')
        if not self.__newfiles.has_key(path):
            self.__newfiles[path] = StringIO()

        buf = self.__newfiles[path]
        buf.seek(offset)
        buf.write(data)
        self._dirty(path)
        return len(data)




# encoding: utf-8
"""Stuff to communicate with JottaCloud servers"""
#
# This file is part of jottabox.
#
# jottabox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jottabox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jottabox.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2016 Håvard Gulldahl <havard@gulldahl.no>

import sys, os, os.path, posixpath, logging, collections

log = logging.getLogger(__name__)

import chardet # get this module: pip install chardet

try:
    from xattr import xattr # get this module: pip install xattr
    HAS_XATTR=True
except ImportError: # no xattr installed, not critical because it is optional
    HAS_XATTR=False

import jottalib
from jottalib.JFS import JFSNotFoundError, \
                         JFSFolder, JFSFile, JFSIncompleteFile, JFSFileDirList, \
                         calculate_md5


#A namedtuple to keep a link between a local path and its online counterpart
#localpath will be a byte string with utf8 code points
#jottapath will be a unicode string
SyncFile = collections.namedtuple('SyncFile', 'localpath, jottapath')

def sf(f, dirpath, jottapath):
    """Create and return a SyncFile tuple from filename.

            localpath will be a byte string with utf8 code points
            jottapath will be a unicode string"""
    log.debug('Create SyncFile from %s', repr(f))
    log.debug('Got encoded filename %r, joining with dirpath %r', _encode_filename_to_filesystem(f), dirpath)
    return SyncFile(localpath=os.path.join(dirpath, _encode_filename_to_filesystem(f)),
                  jottapath=posixpath.join(_decode_filename_to_unicode(jottapath), _decode_filename_to_unicode(f)))


def get_jottapath(localtopdir, dirpath, jottamountpoint):
    """Translate localtopdir to jottapath. Returns unicode string"""
    log.debug("get_jottapath %r %r %r", localtopdir, dirpath, jottamountpoint)
    normpath =  posixpath.normpath(posixpath.join(jottamountpoint, posixpath.basename(localtopdir),
                                   posixpath.relpath(dirpath, localtopdir)))
    return _decode_filename_to_unicode(normpath)


def is_file(jottapath, JFS):
    """Check if a file exists on jottacloud"""
    log.debug("is_file %r", jottapath)
    try:
        jf = JFS.getObject(jottapath)
    except JFSNotFoundError:
        return False
    return isinstance(jf, JFSFile)

def filelist(jottapath, JFS):
    """Get a set() of files from a jottapath (a folder)"""
    log.debug("filelist %r", jottapath)
    try:
        jf = JFS.getObject(jottapath)
    except JFSNotFoundError:
        return set() # folder does not exist, so pretend it is an empty folder
    if not isinstance(jf, JFSFolder):
        return False
    return set([f.name for f in jf.files() if not f.is_deleted()]) # Only return files that aren't deleted

def folderlist(jottapath, JFS):
    """Get a set() of folders from a jottapath (a folder)"""
    logging.debug("folderlist %r", jottapath)
    try:
        jf = JFS.getObject(jottapath)
    except JFSNotFoundError:
        return set() # folder does not exist, so pretend it is an empty folder
    if not isinstance(jf, JFSFolder):
        return False
    return set([f.name for f in jf.folders() if not f.is_deleted()]) # Only return files that aren't deleted

def compare(localtopdir, jottamountpoint, JFS, followlinks=False, exclude_patterns=None):
    """Make a tree of local files and folders and compare it with what's currently on JottaCloud.

    For each folder, yields:
        dirpath, # byte string, full path
        onlylocal, # set(), files that only exist locally, i.e. newly added files that don't exist online,
        onlyremote, # set(), files that only exist in the JottaCloud, i.e. deleted locally
        bothplaces # set(), files that exist both locally and remotely
        onlyremotefolders, # set(), folders that only exist in the JottaCloud, i.e. deleted locally
    """
    def excluded(unicodepath, fname):
        fpath = os.path.join(unicodepath, _decode_filename_to_unicode(fname))
        if exclude_patterns is None:
            return False
        for p in exclude_patterns:
            if p.search(fpath):
                log.debug("%r excluded by pattern %r", fpath, p.pattern)
                return True
        return False
    bytestring_localtopdir = _encode_filename_to_filesystem(localtopdir)
    for dirpath, dirnames, filenames in os.walk(bytestring_localtopdir, followlinks=followlinks):
        # to keep things explicit, and avoid encoding/decoding issues,
        # keep a bytestring AND a unicode variant of dirpath
        dirpath = _encode_filename_to_filesystem(dirpath)
        unicodepath = _decode_filename_to_unicode(dirpath)
        log.debug("compare walk: %r -> %s files ", unicodepath, len(filenames))

        # create set()s of local files and folders
        # paths will be unicode strings
        localfiles = set([f for f in filenames if not excluded(unicodepath, f)]) # these are on local disk
        localfolders = set([f for f in dirnames if not excluded(unicodepath, f)]) # these are on local disk
        jottapath = get_jottapath(localtopdir, unicodepath, jottamountpoint) # translate to jottapath
        log.debug("compare jottapath: %r", jottapath)

        # create set()s of remote files and folders
        # paths will be unicode strings
        cloudfiles = filelist(jottapath, JFS) # set(). these are on jottacloud
        cloudfolders = folderlist(jottapath, JFS)

        log.debug("--cloudfiles: %r", cloudfiles)
        log.debug("--localfiles: %r", localfiles)
        log.debug("--cloudfolders: %r", cloudfolders)

        onlylocal = [ sf(f, dirpath, jottapath) for f in localfiles.difference(cloudfiles)]
        onlyremote = [ sf(f, dirpath, jottapath) for f in cloudfiles.difference(localfiles)]
        bothplaces = [ sf(f, dirpath, jottapath) for f in localfiles.intersection(cloudfiles)]
        onlyremotefolders = [ sf(f, dirpath, jottapath) for f in cloudfolders.difference(localfolders)]
        yield dirpath, onlylocal, onlyremote, bothplaces, onlyremotefolders


def _decode_filename_to_unicode(f):
    '''Get bytestring filename and return unicode.
    First, try to decode from default file system encoding
    If that fails, use ``chardet`` module to guess encoding.
    As a last resort, try to decode as utf-8.

    If the argument already is unicode, return as is'''

    log.debug('_decode_filename_to_unicode(%s)', repr(f))
    if isinstance(f, unicode):
        return f
    try:
        return f.decode(sys.getfilesystemencoding())
    except UnicodeDecodeError:
        charguess = chardet.detect(f)
        log.debug("chardet filename: %r -> %r", f, charguess)
        if charguess['encoding'] is not None:
            try:
                return f.decode(charguess['encoding'])
            except UnicodeDecodeError:
                pass
        log.warning('Cannot understand decoding of this filename: %r (guessed %r, but was wrong)',
                    f, charguess)
        log.debug('Trying utf-8 to decode %r', f)
        try:
            return f.decode('utf-8')
        except UnicodeDecodeError:
            pass
        log.debug('Trying latin1 to decode %r', f)
        try:
            return f.decode('latin1')
        except UnicodeDecodeError:
            log.warning('Exhausted all options. Decoding %r to safe ascii', f)
            return f.decode('ascii', errors='ignore')


def _encode_filename_to_filesystem(f):
    '''Get a unicode filename and return bytestring, encoded to file system default.

    If the argument already is a bytestring, return as is'''
    log.debug('_encode_filename_to_filesystem(%s)', repr(f))
    if isinstance(f, str):
        return f
    try:
        return f.encode(sys.getfilesystemencoding())
    except UnicodeEncodeError:
        raise

def new(localfile, jottapath, JFS):
    """Upload a new file from local disk (doesn't exist on JottaCloud).

    Returns JottaFile object"""
    with open(localfile) as lf:
        _new = JFS.up(jottapath, lf)
    return _new

def resume(localfile, jottafile, JFS):
    """Continue uploading a new file from local file (already exists on JottaCloud"""
    with open(localfile) as lf:
        _complete = jottafile.resume(lf)
    return _complete

def replace_if_changed(localfile, jottapath, JFS):
    """Compare md5 hash to determine if contents have changed.
    Upload a file from local disk and replace file on JottaCloud if the md5s differ,
    or continue uploading if the file is incompletely uploaded.

    Returns the JottaFile object"""
    jf = JFS.getObject(jottapath)
    lf_hash = getxattrhash(localfile) # try to read previous hash, stored in xattr
    if lf_hash is None:               # no valid hash found in xattr,
        with open(localfile) as lf:
            lf_hash = calculate_md5(lf) # (re)calculate it
    if type(jf) == JFSIncompleteFile:
        log.debug("Local file %s is incompletely uploaded, continue", localfile)
        return resume(localfile, jf, JFS)
    elif jf.md5 == lf_hash: # hashes are the same
        log.debug("hash match (%s), file contents haven't changed", lf_hash)
        setxattrhash(localfile, lf_hash)
        return jf         # return the version from jottaclouds
    else:
        setxattrhash(localfile, lf_hash)
        return new(localfile, jottapath, JFS)

def deleteDir(jottapath, JFS):
    """Remove folder from JottaCloud because it is no longer present on local disk.
    Returns boolean"""
    jf = JFS.post('%s?dlDir=true' % jottapath)
    return jf.is_deleted()

def delete(jottapath, JFS):
    """Remove file from JottaCloud because it is no longer present on local disk.
    Returns boolean"""
    jf = JFS.post('%s?dl=true' % jottapath)
    return jf.is_deleted()

def mkdir(jottapath, JFS):
    """Make a new directory (a.k.a. folder) on JottaCloud.
    Returns boolean"""
    jf = JFS.post('%s?mkDir=true' % jottapath)
    return instanceof(jf, JFSFolder)

def iter_tree(jottapath, JFS):
    """Get a tree of of files and folders. use as an iterator, you get something like os.walk"""
    filedirlist = JFS.getObject('%s?mode=list' % jottapath)
    log.debug("got tree: %s", filedirlist)
    if not isinstance(filedirlist, JFSFileDirList):
        yield ( '', tuple(), tuple() )
    for path in filedirlist.tree:
        yield path

def setxattrhash(filename, md5hash):
    log.debug('set xattr hash for %s', filename)
    if not HAS_XATTR:
        log.info("xattr not found. Installing it will speed up hash comparison: pip install xattr")
        return False
    try:
        x = xattr(filename)
        x.set('user.jottalib.md5', md5hash)
        x.set('user.jottalib.timestamp', str(os.path.getmtime(filename)))
        x.set('user.jottalib.filesize', str(os.path.getsize(filename)))
        return True
    except IOError:
        # no file system support
        return False
    except Exception as e:
        log.exception(e)
        #log.debug('setxattr got exception %r', e)
    return False

def getxattrhash(filename):
    log.debug('get xattr hash for %s', filename)
    if not HAS_XATTR:
        log.info("xattr not found. Installing it will speed up hash comparison: pip install xattr")
        return None
    try:
        x = xattr(filename)
        if x.get('user.jottalib.filesize') != str(os.path.getsize(filename)) or x.get('user.jottalib.timestamp') != str(os.path.getmtime(filename)):
            x.remove('user.jottalib.filesize')
            x.remove('user.jottalib.timestamp')
            x.remove('user.jottalib.md5')
            return None # this is not the file we have calculated md5 for
        return x.get('user.jottalib.md5')
    except Exception as e:
        log.debug('setxattr got exception %r', e)
        return None



# -*- encoding: utf-8 -*-
#
# This file is part of jottafs.
#
# jottafs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jottafs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jottafs.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2011,2013-2016 Håvard Gulldahl <havard@gulldahl.no>

from __future__ import absolute_import, division, unicode_literals

__author__ = 'havard@gulldahl.no'

import argparse
import six
from six.moves import http_client
import humanize as _humanize
import logging
import os
import posixpath
import sys
import time
import re
from clint.textui import progress
from functools import partial

# import our stuff
from jottalib import JFS, __version__
from .scanner import filescanner

HAS_FUSE = False
try:
    from fuse import FUSE # pylint: disable=unused-import
    HAS_FUSE = True
except ImportError:
    pass

HAS_WATCHDOG = False # for monitor()
try:
    import watchdog
    HAS_WATCHDOG = True
except ImportError:
    pass

if sys.platform != "win32":
    # change progress indicators to something that looks nice
    #TODO: rather detect utf-8 support in the terminal
    ProgressBar = partial(progress.Bar, empty_char=u'○', filled_char=u'●')
else:
    ProgressBar = partial(progress.Bar)


## HELPER FUNCTIONS ##

def get_jotta_device(jfs):
    jottadev = None
    for j in jfs.devices: # find Jotta/Shared folder
        if j.name == 'Jotta':
            jottadev = j
    return jottadev


def get_root_dir(jfs):
    jottadev = get_jotta_device(jfs)
    root_dir = jottadev.mountPoints['Sync']
    return root_dir

def parse_args_and_apply_logging_level(parser, argv):
    args = parser.parse_args(argv)
    logging.basicConfig(level=getattr(logging, args.loglevel.upper()))
    logging.captureWarnings(True)
    http_client.HTTPConnection.debuglevel = 1 if args.loglevel == 'debug' else 0
    return args


def print_size(num, humanize=False):
    if humanize:
        return _humanize.naturalsize(num, gnu=True)
    else:
        return str(num)

def commandline_text(bytestring):
    'Convert bytestring from command line to unicode, using default file system encoding'
    unicode_string = bytestring.decode(sys.getfilesystemencoding())
    return unicode_string

def is_dir(path):
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError('%s is not a valid directory' % path)
    return path.decode(sys.getfilesystemencoding())

## UTILITIES, ONE PER FUNCTION ##


def fuse(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    if not HAS_FUSE:
        message = ['jotta-fuse requires fusepy (pip install fusepy), install that and try again.']
        if os.name == 'nt':
            message.append('Note: jotta-fuse is not supported on Windows, but Cygwin might work.')
        print(' '.join(message))
        sys.exit(1)


    from .jottafuse import JottaFuse
    parser = argparse.ArgumentParser(description=__doc__,
                                     epilog="""The program expects to find an entry for "jottacloud.com" in your .netrc,
                                     or JOTTACLOUD_USERNAME and JOTTACLOUD_PASSWORD in the running environment.
                                     This is not an official JottaCloud project.""")
    parser.add_argument('--debug',
                        action='store_true',
                        help='Run fuse in the foreground and add a lot of messages to help debug')
    parser.add_argument('--debug-fuse',
                        action='store_true',
                        help='Show all low-level filesystem operations')
    parser.add_argument('--debug-http',
                        action='store_true',
                        help='Show all HTTP traffic')
    parser.add_argument('--version',
                        action='version', version=__version__)
    parser.add_argument('mountpoint',
                        type=is_dir,
                        help='A path to an existing directory where you want your JottaCloud tree mounted')
    args = parser.parse_args(argv)
    if args.debug_http:
        http_client.HTTPConnection.debuglevel = 1
    if args.debug:
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True
        logging.basicConfig(level=logging.DEBUG)

    auth = JFS.get_auth_info()
    fuse = FUSE(JottaFuse(auth), args.mountpoint, debug=args.debug_fuse,
                sync_read=True, foreground=args.debug, raw_fi=False,
                fsname="JottaCloudFS", subtype="fuse")

def upload(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Upload a file to JottaCloud.')
    parser.add_argument('localfile',
                        help='The local file that you want to upload',
                        type=argparse.FileType('r'))
    parser.add_argument('remote_dir',
                        help='The remote directory to upload the file to',
                        nargs='?',
                        type=commandline_text)
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    jfs = JFS.JFS()
    args = parse_args_and_apply_logging_level(parser, argv)
    decoded_filename = commandline_text(args.localfile.name)
    progress_bar = ProgressBar()
    callback = lambda monitor, size: progress_bar.show(monitor.bytes_read, size)
    root_folder = get_root_dir(jfs)
    if args.remote_dir:
        target_dir_path = posixpath.join(root_folder.path, args.remote_dir)
        target_dir = jfs.getObject(target_dir_path)
    else:
        target_dir = root_folder
    upload = target_dir.up(args.localfile, os.path.basename(decoded_filename), upload_callback=callback)
    print('%s uploaded successfully' % decoded_filename)
    return True # TODO: check return value


def share(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Share a file on JottaCloud and get the public URI.',
                                     epilog='Note: This utility needs to find JOTTACLOUD_USERNAME'
                                     ' and JOTTACLOUD_PASSWORD in the running environment.')
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    parser.add_argument('localfile',
                        help='The local file that you want to share',
                        type=argparse.FileType('r'))
    args = parse_args_and_apply_logging_level(parser, argv)
    jfs = JFS.JFS()
    jottadev = get_jotta_device(jfs)
    jottashare = jottadev.mountPoints['Shared']
    upload = jottashare.up(args.localfile)  # upload file
    public = upload.share() # share file
    logging.debug('Shared %r and got: %r (%s)', args.localfile, public, dir(public))
    for (filename, uuid, publicURI) in public.sharedFiles():
        print('%s is now available to the world at %s' % (filename, publicURI))
    return True # TODO: check return value of command


def ls(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='List files in Jotta folder.', add_help=False)
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    parser.add_argument('-h', '--humanize', # this matches ls(1)
                        help='Print human-readable file sizes.',
                        action='store_true')
    parser.add_argument('-a',
                        '--all',
                        action='store_true',
                        help='Include deleted and incomplete files (otherwise ignored)')
    parser.add_argument('item',
                        nargs='?',
                        help='The file or directory to list. Defaults to the root dir',
                        type=commandline_text)
    parser.add_argument('-H', # because -h means --humanize
                        '--help',
                        help='Print this help',
                        action='help')
    args = parse_args_and_apply_logging_level(parser, argv)
    jfs = JFS.JFS()
    root_folder = get_root_dir(jfs)
    if args.item:
        if args.item.startswith('//'):
            # break out of root_folder
            item_path = posixpath.join(jfs.rootpath, args.item[1:])
        else:
            item_path = posixpath.join(root_folder.path, args.item)
        item = jfs.getObject(item_path)
    else:
        item = root_folder
    timestamp_width = 25
    logging.debug('about to ls %r', item)
    if isinstance(item, JFS.JFSFolder):
        files = [(
            f.created,
            print_size(f.size, humanize=args.humanize) if f.size else u'',
            u'D' if f.deleted else u'I' if f.state == 'INCOMPLETE' else u' ',
            f.name) for f in item.files() if not f.deleted and f.state != 'INCOMPLETE' or args.all]
        folders = [(u' '*timestamp_width, u'', u'D' if f.deleted else u' ', unicode(f.name))
                   for f in item.folders() if not f.deleted or args.all]
        widest_size = 0
        for f in files:
            if len(f[1]) > widest_size:
                widest_size = len(f[1])
        for item in sorted(files + folders, key=lambda t: t[3]):
            if args.all:
                print(u'%s %s %s %s' % (item[0], item[1].rjust(widest_size), item[2], item[3]))
            else:
                print(u'%s %s %s' % (item[0], item[1].rjust(widest_size), item[3]))
    else:
        print(' '.join([str(item.created), print_size(item.size, humanize=args.humanize), item.name]))
    return True # TODO: check return value of command


def download(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Download a file from Jottacloud.')
    parser.add_argument('remotefile',
                        help='The path to the file that you want to download',
                        type=commandline_text)
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    args = parse_args_and_apply_logging_level(parser, argv)
    jfs = JFS.JFS()
    root_folder = get_root_dir(jfs)
    path_to_object = posixpath.join(root_folder.path, args.remotefile)
    remote_file = jfs.getObject(path_to_object)
    total_size = remote_file.size
    with open(remote_file.name, 'wb') as fh:
        bytes_read = 0
        with ProgressBar(expected_size=total_size) as bar:
            for chunk_num, chunk in enumerate(remote_file.stream()):
                fh.write(chunk)
                bytes_read += len(chunk)
                bar.show(bytes_read)
    print('%s downloaded successfully' % args.remotefile)
    return True # TODO: check/set return value


def mkdir(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Create a new folder in Jottacloud.')
    parser.add_argument('newdir',
                        help='The path to the folder that you want to create')
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    args = parse_args_and_apply_logging_level(parser, argv)
    jfs = JFS.JFS()
    root_folder = get_root_dir(jfs)
    root_folder.mkdir(args.newdir)
    return True #  TODO: check return value of mkdir


def rm(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Delete an item from Jottacloud')
    parser.add_argument('file',
                        help='The path to the item that you want to delete')
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    parser.add_argument('-f', '--force',
                        help='Completely deleted, no restore possiblity',
                        action='store_true')
    args = parse_args_and_apply_logging_level(parser, argv)
    jfs = JFS.JFS()
    root_dir = get_root_dir(jfs)
    item_path = posixpath.join(root_dir.path, args.file)
    item = jfs.getObject(item_path)
    if args.force:
        item.hard_delete()
    else:
        item.delete()
    print('%s deleted' % args.file)
    return True # TODO: check return value of command


def restore(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Restore a deleted item from Jottacloud')
    parser.add_argument('file',
                        type=commandline_text,
                        help='The path to the item that you want to restore')
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    args = parse_args_and_apply_logging_level(parser, argv)
    jfs = JFS.JFS()
    root_dir = get_root_dir(jfs)
    item_path = posixpath.join(root_dir.path, args.file)
    item = jfs.getObject(item_path)
    item.restore()
    print('%s restored' % args.file)
    return True # TODO: check return value of command



def cat(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Display contents of a file from Jottacloud')
    parser.add_argument('file',
                        type=commandline_text,
                        help='The path to the file that you want to show')
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    args = parse_args_and_apply_logging_level(parser, argv)
    jfs = JFS.JFS()
    if args.file.startswith('//'):
        # break out of root_folder
        item_path = posixpath.join(jfs.rootpath, args.file[1:])
    else:
        root_dir = get_root_dir(jfs)
        item_path = posixpath.join(root_dir.path, args.file)
    item = jfs.getObject(item_path)
    if not isinstance(item, JFS.JFSFile):
        print("%r is not a file (it's a %s), so we can't show it" % (args.file, type(item)))
        sys.exit(1)
    s = ''
    for chunk in item.stream():
        print(chunk.encode(sys.getdefaultencoding()))
        s = s + chunk
    return s


def scanner(argv=None):

    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(description=__doc__,
                                    epilog="""The program expects to find an entry for "jottacloud.com" in your .netrc,
                                    or JOTTACLOUD_USERNAME and JOTTACLOUD_PASSWORD in the running environment.
                                    This is not an official JottaCloud project.""")
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    parser.add_argument('--errorfile',
                        type=commandline_text,
                        help='A file to write errors to',
                        default='./jottacloudclient.log')
    parser.add_argument('--exclude',
                        type=re.compile,
                        action='append',
                        help='Exclude paths matched by this pattern (can be repeated)')
    parser.add_argument('--prune-files', dest='prune_files',
                        help='Delete files that does not exist locally',
                        action='store_true')
    parser.add_argument('--prune-folders',
                        dest='prune_folders',
                        help='Delete folders that does not exist locally',
                        action='store_true')
    parser.add_argument('--prune-all',
                        dest='prune_all',
                        help='Combines --prune-files  and --prune-folders',
                        action='store_true')
    parser.add_argument('--version',
                        action='version',
                        version=__version__)
    parser.add_argument('--dry-run',
                        action='store_true',
                        help="don't actually do any uploads or deletes, just show what would be done")
    parser.add_argument('topdir',
                        type=is_dir,
                        help='Path to local dir that needs syncing')
    parser.add_argument('jottapath',
                        type=commandline_text,
                        help='The path at JottaCloud where the tree shall be synced (must exist)')
    args = parse_args_and_apply_logging_level(parser, argv)
    if args.prune_all:
        args.prune_files = True
        args.prune_folders = True

    fh = logging.FileHandler(args.errorfile)
    fh.setLevel(logging.ERROR)
    logging.getLogger('').addHandler(fh)

    jfs = JFS.JFS()

    logging.info('args: topdir %r, jottapath %r', args.topdir, args.jottapath)
    filescanner(args.topdir, args.jottapath, jfs, args.errorfile, args.exclude, args.dry_run, args.prune_files, args.prune_folders)


def monitor(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    if not HAS_WATCHDOG:
        message = ['jotta-monitor requires watchdog (pip install watchdog), install that and try again.']
        print(' '.join(message))
        sys.exit(1)

    # Has watchdog, can safely import filemonitor
    from .monitor import filemonitor

    parser = argparse.ArgumentParser(description=__doc__,
                                    epilog="""The program expects to find an entry for "jottacloud.com" in your .netrc,
                                    or JOTTACLOUD_USERNAME and JOTTACLOUD_PASSWORD in the running environment.
                                    This is not an official JottaCloud project.""")
    parser.add_argument('-l', '--loglevel',
                        help='Logging level. Default: %(default)s.',
                        choices=('debug', 'info', 'warning', 'error'),
                        default='warning')
    parser.add_argument('--errorfile',
                        help='A file to write errors to',
                        default='./jottacloudclient.log')
    parser.add_argument('--version',
                        action='version',
                        version=__version__)
    parser.add_argument('--dry-run',
                        action='store_true',
                        help="don't actually do any uploads or deletes, just show what would be done")
    parser.add_argument('topdir',
                        type=is_dir,
                        help='Path to local dir that needs syncing')
    parser.add_argument('mode',
                        type=commandline_text,
                        help='Mode of operation: ARCHIVE, SYNC or SHARE. See README.md',
                        choices=( 'archive', 'sync', 'share') )
    args = parse_args_and_apply_logging_level(parser, argv)
    fh = logging.FileHandler(args.errorfile)
    fh.setLevel(logging.ERROR)
    logging.getLogger('').addHandler(fh)

    jfs = JFS.JFS()

    filemonitor(args.topdir, args.mode, jfs)



import struct
import sys

# .lnk parser by user Jared at StackOverflow http://stackoverflow.com/a/28952464
# retrieved 2015-07-21 by havard@gulldahl.no
# added .decode(sys.getfilesystemencoding()) to handle non-ascii targets
#
# Jared added the following comment to their code:
# I know this is an older thread but I feel that there isn't much information on the method that uses the link specification as noted in the original question.
#
#My shortcut target implementation could not use the win32com module and after a lot of searching, decided to come up with my own. Nothing else seemed to accomplish what I needed under my restrictions. Hopefully this will help other folks in this same situation.
#
#It uses the binary structure Microsoft has provided for MS-SHLLINK.


def readlnk(path):
    target = ''
    try:
        with open(path, 'rb') as stream:
            content = stream.read()

            # skip first 20 bytes (HeaderSize and LinkCLSID)
            # read the LinkFlags structure (4 bytes)
            lflags = struct.unpack('I', content[0x14:0x18])[0]
            position = 0x18

            # if the HasLinkTargetIDList bit is set then skip the stored IDList
            # structure and header
            if (lflags & 0x01) == 1:
                position = struct.unpack('H', content[0x4C:0x4E])[0] + 0x4E

            last_pos = position
            position += 0x04

            # get how long the file information is (LinkInfoSize)
            length = struct.unpack('I', content[last_pos:position])[0]

            # skip 12 bytes (LinkInfoHeaderSize, LinkInfoFlags, and VolumeIDOffset)
            position += 0x0C

            # go to the LocalBasePath position
            lbpos = struct.unpack('I', content[position:position+0x04])[0]
            position = last_pos + lbpos

            # read the string at the given position of the determined length
            size= (length + last_pos) - position - 0x02
            temp = struct.unpack('c' * size, content[position:position+size])
            target = ''.join([chr(ord(a)) for a in temp])
    except:
        # could not read the file
        pass

    return target.decode(sys.getfilesystemencoding())


if __name__ == '__main__':
    print(readlnk(sys.argv[1]))



#
# Created by S W on Sun, 31 Oct 2004 (PSF License)
#
# https://code.activestate.com/recipes/325905-memoize-decorator-with-timeout/
#
# This simple decorator is different to other memoize decorators in that it will
# only cache results for a period of time. It also provides a simple method of cleaning
# the cache of old entries via the .collect method. This will help prevent excessive
# or needless memory consumption.
#
#
# #The code below demonstrates usage of the MWT decorator. Notice how the cache is
# #cleared of some entries after the MWT().collect() method is called.
#
# @MWT()
# def z(a,b):
#     return a + b
#
# @MWT(timeout=5)
# def x(a,b):
#     return a + b
#
# z(1,2)
# x(1,3)
#
#
# print MWT()._caches
# >>> {<function 'z'>: {(1, 2): (3, 1099276281.092)},<function 'x'> : {(1, 3): (4, 1099276281.092)}}
#
# time.sleep(3)
# MWT().collect()
# print MWT()._caches
#>>> {<function 'z'>: {},<function 'x'> : {(1, 3): (4, 1099276281.092)}}
#

import time
import logging

log = logging.getLogger(__name__)

class MWT(object):
    """Memoize With Timeout"""
    _caches = {}
    _timeouts = {}

    def __init__(self,timeout=2):
        self.timeout = timeout

    def collect(self):
        """Clear cache of results which have timed out"""
        for func in self._caches:
            cache = {}
            for key in self._caches[func]:
                if (time.time() - self._caches[func][key][1]) < self._timeouts[func]:
                    cache[key] = self._caches[func][key]
            self._caches[func] = cache

    def __call__(self, f):
        self.cache = self._caches[f] = {}
        self._timeouts[f] = self.timeout

        def func(*args, **kwargs):
            kw = kwargs.items()
            kw.sort()
            key = (args, tuple(kw))
            try:
                v = self.cache[key]
                log.debug("get object from cache: %s", key)
                if (time.time() - v[1]) > self.timeout:
                    raise KeyError
            except KeyError:
                log.debug("new object in cache: %s", key)
                v = self.cache[key] = f(*args,**kwargs),time.time()
            return v[0]
        func.func_name = f.func_name

        return func


#
#
# This stuff added by the jottalib project, under the same terms as above (PSF License)
#

class Memoize(MWT):
    '''A superset of MWT that adds the possibility to yank paths from cached results'''
    def yank_path(self, path):
        """Clear cache of results from a specific path"""
        for func in self._caches:
            cache = {}
            for key in self._caches[func].keys():
                log.debug("cache key %s for func %s", key, func)
                if path in key[0]:
                    log.debug("del cache key %s", key)
                    del self._caches[func][key]



# -*- encoding: utf-8 -*-
#
# This file is part of jottafs.
#
# jottafs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jottafs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jottafs.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2011,2013,2014 Håvard Gulldahl <havard@gulldahl.no>

# metadata

__author__ = 'havard@gulldahl.no'
from jottalib import __version__

# importing stdlib
import sys, os, os.path, time
import posixpath, logging, datetime, hashlib
from collections import namedtuple
import six
from six.moves import cStringIO as StringIO

# importing external dependencies (pip these, please!)
import requests
from requests.utils import quote
import netrc
import requests_toolbelt
import certifi

import lxml, lxml.objectify
import dateutil, dateutil.parser # pip install python-dateutil

log = logging.getLogger(__name__)

#monkeypatch urllib3 param function to bypass bug in jottacloud servers
from requests.packages import urllib3
urllib3.fields.format_header_param_orig = urllib3.fields.format_header_param
def mp(name, value):
    return urllib3.fields.format_header_param_orig(name, value).replace('filename*=', 'filename=')
urllib3.fields.format_header_param = mp

# some setup
JFS_ROOT='https://www.jottacloud.com/jfs/'

# helper functions
try:
    unicode("we are python2")
except NameError:
    def unicode(s): return str(s) # TODO: use six
##
# A note regarding the use of unicode() to convert string values from XML object nodes (lxml.objectify)
#
# you are always safe to convert objectify.StringElements to native Python unicode objects using unicode()
# - https://mailman-mail5.webfaction.com/pipermail/lxml/2011-February/005885.html
#


def get_auth_info():
    """ Get authentication details to jottacloud.

    Will first check environment variables, then the .netrc file.
    """
    env_username = os.environ.get('JOTTACLOUD_USERNAME')
    env_password = os.environ.get('JOTTACLOUD_PASSWORD')
    netrc_auth = None
    try:
        netrc_file = netrc.netrc()
        netrc_auth = netrc_file.authenticators('jottacloud.com')
    except IOError:
        # .netrc file doesn't exist
        pass
    netrc_username = None
    netrc_password = None
    if netrc_auth:
        netrc_username, _, netrc_password = netrc_auth
    username = env_username or netrc_username
    password = env_password or netrc_password
    if not (username and password):
        raise JFSError('Could not find username and password in either env or ~/.netrc, '
            'you need to add one of these to use these tools')
    return (username, password)

def calculate_md5(fileobject, size=2**16):
    """Utility function to calculate md5 hashes while being light on memory usage.

    By reading the fileobject piece by piece, we are able to process content that
    is larger than available memory"""
    fileobject.seek(0)
    md5 = hashlib.md5()
    for data in iter(lambda: fileobject.read(size), b''):
        if isinstance(data, unicode):
            data = data.encode('utf-8') # md5 needs a byte string
        md5.update(data)
    fileobject.seek(0) # rewind read head
    return md5.hexdigest()

# error classes

class JFSError(Exception):
    @staticmethod
    def raiseError(e, path): # parse object from lxml.objectify and
        if(e.code) == 404:
            raise JFSNotFoundError('%s does not exist (%s)' % (path, e.message))
        elif(e.code) == 401:
            raise JFSCredentialsError("Your credentials don't match for %s (%s) (probably incorrect password!)" % (path, e.message))
        elif(e.code) == 403:
            raise JFSAuthenticationError("You don't have access to %s (%s)" % (path, e.message))
        elif(e.code) == 416:
            raise JFSRangeError("Requested Range Not Satisfiable (%s)" % e.message)
        elif(e.code) == 500:
            raise JFSServerError("Internal server error: %s (%s)" % (path, e.message))
        elif(e.code) == 400:
            raise JFSBadRequestError('Bad request: %s (%s)' % (path, e.message))
        else:
            raise JFSError('Error accessing %s (%s)' % (path, e.message))

class JFSBadRequestError(JFSError): # HTTP 400
    pass

class JFSCredentialsError(JFSError): # HTTP 401
    pass

class JFSNotFoundError(JFSError): # HTTP 404
    pass

class JFSAccessError(JFSError): #
    pass

class JFSAuthenticationError(JFSError): # HTTP 403
    pass

class JFSRangeError(JFSError): # HTTP 416
    pass

class JFSServerError(JFSError): # HTTP 500
    pass

# classes mapping JFS structures


class JFSFileDirList(object):
    '''Wrapping <filedirlist>, a simple tree of folders and their files

       Get a <filedirlist> for any jottafolder by appending ?mode=list to your query

       Then you will get this object, with a .tree property, which is a list of all
       files and folders.

       Files will be a namedtuple() with five properties:
         .name - file name
         .state - file state (str): one of JFS.ProtoFile.STATE_*, e.g 'COMPLETED' or 'INCOMPLETE'
         .size - file size (int or None): full size , partially uploaded size or no size, depentant on file state
         .md5 - jottacloud file hash (str or None): corrupt files have no md5 hash
         .uuid - jottacloud assigned uuid

    <filedirlist time="2015-05-28-T18:57:06Z" host="dn-093.site-000.jotta.no">
        <folders>
          <folder name="Sync">
            <path xml:space="preserve">/havardgulldahl/Jotta</path>
            <abspath xml:space="preserve">/havardgulldahl/Jotta</abspath>
            <files>
                <file>...'''


    def __init__(self, filedirlistobject, jfs, parentpath): # filedirlistobject from lxml.objectify
        self.filedirlist = filedirlistobject
        self.parentPath = parentpath
        self.jfs = jfs

        treefile = namedtuple('TreeFile', 'name size md5 uuid state')

        self.tree = {}
        for folder in self.filedirlist.folders.iterchildren():
            foldername = unicode(folder.attrib.get('name'))
            path = unicode(folder.path)
            t = []
            if hasattr(folder, 'files'):
                for file_ in folder.files.iterchildren():
                    if hasattr(file_, 'currentRevision'): # a normal file
                        t.append(treefile(unicode(file_.attrib['name']),
                                          int(file_.currentRevision.size),
                                          unicode(file_.currentRevision.md5),
                                          unicode(file_.attrib['uuid']),
                                          unicode(file_.currentRevision.state)
                                          )
                                 )
                    else:
                        # This is an incomplete or, possibly, corrupt file
                        #
                        # Incomplete files have no `size` in a filedirlist, you
                        # need to fetch the JFSFile explicitly to see that property
                        try:
                            # incomplete files carry a md5 hash,
                            _md5 = unicode(file_.latestRevision.md5)
                        except AttributeError:
                            # while other may not
                            # see discussion in #88
                            _md5 = None
                        t.append(treefile(unicode(file_.attrib['name']),
                                          None, # return size as None
                                          _md5,
                                          unicode(file_.attrib['uuid']),
                                          unicode(file_.latestRevision.state)
                                          )
                                 )
            self.tree[posixpath.join(path, foldername)] = t



class JFSFolder(object):
    'OO interface to a folder, for convenient access. Type less, do more.'
    def __init__(self, folderobject, jfs, parentpath): # folderobject from lxml.objectify
        self.folder = folderobject
        self.parentPath = parentpath
        self.jfs = jfs
        self.synced = False

    @property
    def name(self):
        if self.folder.attrib.has_key('name'):
            return unicode(self.folder.attrib['name'])
        return unicode(self.folder.name)

    @property
    def path(self):
        log.debug('path: %r + %r', self.parentPath, self.name)
        return '%s/%s' % (self.parentPath, self.name)

    @property
    def deleted(self):
        'Return datetime.datetime or None if the file isnt deleted'
        _d = self.folder.attrib.get('deleted', None)
        if _d is None: return None
        return dateutil.parser.parse(str(_d))

    def sync(self):
        'Update state of folder from Jottacloud server'
        log.info("syncing %r" % self.path)
        self.folder = self.jfs.get(self.path)
        self.synced = True

    def is_deleted(self):
        'Return bool based on self.deleted'
        return self.deleted is not None

    def files(self):
        if not self.synced:
            self.sync()
        try:
            #return [JFSFile(f, self.jfs, self.path) for f in self.folder.files.iterchildren()]
            for _f in self.folder.files.iterchildren():
                if hasattr(_f, 'currentRevision'): # a normal file
                    yield JFSFile(_f, self.jfs, self.path)
                else:
                    yield JFSIncompleteFile(_f, self.jfs, self.path)

        except AttributeError:
            while False:
                yield None
            #return [x for x in []]

    def folders(self):
        if not self.synced:
            self.sync()
        try:
            return [JFSFolder(f, self.jfs, self.path) for f in self.folder.folders.iterchildren()]
        except AttributeError:
            return [x for x in []]

    def mkdir(self, foldername):
        'Create a new subfolder and return the new JFSFolder'
        #url = '%s?mkDir=true' % posixpath.join(self.path, foldername)
        url = posixpath.join(self.path, foldername)
        params = {'mkDir':'true'}
        r = self.jfs.post(url, params)
        self.sync()
        return r

    def restore(self):
        'Restore the folder'
        #
        #
        # As of 2016-06-15, Jottacloud.com has changed their restore api
        # To restore, this is what's done
        #
        # HTTP POST to https://www.jottacloud.com/web/restore/trash/list
        # Data:
        #     hash:undefined
        #     files:@0025d37be5329a18eece18dd93f793509e8_dGVzdF9kZWxldGUudHh0
        #
        # where `files` is a comma separated list, and each item is constructed thus:
        #  @<uuid of path>_<base64 encoded file name>
        #
        if not self.deleted:
            raise JFSError('Tried to restore a not deleted folder')
        raise NotImplementedError
        url = 'https://www.jottacloud.com/rest/webrest/%s/action/restore' % self.jfs.username
        data = {'paths[]': self.path.replace(JFS_ROOT, ''),
                'web': 'true',
                'ts': int(time.time()),
                'authToken': 0}
        r = self.jfs.post(url, content=data)
        return r

    def delete(self):
        'Delete this folder and return a deleted JFSFolder'
        #url = '%s?dlDir=true' % self.path
        params = {'dlDir':'true'}
        r = self.jfs.post(self.path, params)
        self.sync()
        return r

    def hard_delete(self):
        'Deletes without possibility to restore'
        url = 'https://www.jottacloud.com/rest/webrest/%s/action/delete' % self.jfs.username
        data = {'paths[]': self.path.replace(JFS_ROOT, ''),
                'web': 'true',
                'ts': int(time.time()),
                'authToken': 0}
        r = self.jfs.post(url, content=data)
        return r


    def rename(self, newpath):
        "Move folder to a new name, possibly a whole new path"
        # POST https://www.jottacloud.com/jfs/**USERNAME**/Jotta/Sync/Ny%20mappe?mvDir=/**USERNAME**/Jotta/Sync/testFolder
        #url = '%s?mvDir=/%s%s' % (self.path, self.jfs.username, newpath)
        params = {'mvDir':'/%s%s' % (self.jfs.username, newpath)}
        r = self.jfs.post(self.path,
                          extra_headers={'Content-Type':'application/octet-stream'},
                          params=params)
        return r

    def up(self, fileobj_or_path, filename=None, upload_callback=None):
        'Upload a file to current folder and return the new JFSFile'
        close_on_done = False

        if isinstance(fileobj_or_path, six.string_types):
            filename = filename or os.path.basename(fileobj_or_path)
            fileobj_or_path = open(fileobj_or_path, 'rb')
            close_on_done = True
        elif hasattr(fileobj_or_path, 'read'):  # file like
            pass
        else:
            # TODO: handle generators here?
            raise JFSError("Need filename or file-like object")

        if filename is None:
            if hasattr(fileobj_or_path, 'name'):
                filename = os.path.basename(fileobj_or_path.name)
            else:
                raise JFSError("Unable to guess filename")

        log.debug('.up %s ->  %s %s', repr(fileobj_or_path), repr(self.path), repr(filename))
        r = self.jfs.up(posixpath.join(self.path, filename), fileobj_or_path,
            upload_callback=upload_callback)
        if close_on_done:
            fileobj_or_path.close()
        self.sync()
        return r

    def filedirlist(self):
        'Get a JFSFileDirList, recursive tree of JFSFile and JFSFolder'
        #url = '%s?mode=list' % self.path
        params = {'mode':'list'}
        return self.jfs.getObject(self.path, params)

class ProtoFile(object):
    'Prototype for different incarnations fo file, e.g. JFSIncompleteFile and JFSFile'

    # constants for known file states
    STATE_COMPLETED = 'COMPLETED' # -> JFSFile
    STATE_ADDED = 'ADDED'
    STATE_INCOMPLETE = 'INCOMPLETE' # -> JFSIncompleteFile
    STATE_PROCESSING = 'PROCESSING'
    STATE_CORRUPT = 'CORRUPT' # -> JFSCorruptFile
    @staticmethod
    def factory(fileobject, jfs, parentpath): # fileobject from lxml.objectify
        'Class method to get the correct file class instatiated'
        if hasattr(fileobject, 'currentRevision'): # a normal file
            return JFSFile(fileobject, jfs, parentpath)
        elif str(fileobject.latestRevision.state) == ProtoFile.STATE_INCOMPLETE:
            return JFSIncompleteFile(fileobject, jfs, parentpath)
        elif str(fileobject.latestRevision.state) == ProtoFile.STATE_CORRUPT:
            return JFSCorruptFile(fileobject, jfs, parentpath)
        else:
            raise NotImplementedError('No JFS*File support for state %r. Please file a bug!' % fileobject.latestRevision.state)

    def __init__(self, fileobject, jfs, parentpath): # fileobject from lxml.objectify
        self.f = fileobject
        self.jfs = jfs
        self.parentPath = parentpath

    def is_image(self):
        'Return bool based on self.mime'
        return os.path.dirname(self.mime) == 'image'

    @property
    def name(self):
        return unicode(self.f.attrib['name'])

    @property
    def uuid(self):
        return unicode(self.f.attrib['uuid'])

    @property
    def deleted(self):
        'Return datetime.datetime or None if the file isnt deleted'
        _d = self.f.attrib.get('deleted', None)
        if _d is None: return None
        return dateutil.parser.parse(str(_d))

    def is_deleted(self):
        'Return bool based on self.deleted'
        return self.deleted is not None

    @property
    def path(self):
        return posixpath.join(self.parentPath, self.name)


class JFSCorruptFile(ProtoFile):
    'OO interface to a corrupt file.'
    """
 18     <revision>
 19       <number>3</number>
 20       <state>CORRUPT</state>
 21       <created>2016-06-14-T19:09:47Z</created>
 22       <modified>2016-06-14-T19:09:47Z</modified>
 23       <mime>text/plain</mime>
 24       <mstyle>text/plain</mstyle>
 25       <md5>2ed82c2b9a78f3fce85b19592fc94581</md5>
 26       <updated>2016-06-14-T19:09:48Z</updated>
 27     </revision>"""

    @property
    def revisionNumber(self):
        'return int of current revision'
        return int(self.f.latestRevision.number)

    @property
    def created(self):
        'return datetime.datetime'
        return dateutil.parser.parse(str(self.f.latestRevision.created))

    @property
    def modified(self):
        'return datetime.datetime'
        return dateutil.parser.parse(str(self.f.latestRevision.modified))

    @property
    def updated(self):
        'return datetime.datetime'
        return dateutil.parser.parse(str(self.f.latestRevision.updated))

    @property
    def md5(self):
        return str(self.f.latestRevision.md5)

    @property
    def mime(self):
        return unicode(self.f.latestRevision.mime)

    @property
    def state(self):
        return unicode(self.f.latestRevision.state)

class JFSIncompleteFile(JFSCorruptFile):
    'OO interface to an incomplete file. Like JFSCorruptFile, but adds a .size property and a .resume() method.'
    """<file name="iii.m4v" uuid="e8f268ac-d081-4d4f-bfb1-77149b2bd51d" time="2015-05-29-T18:11:56Z" host="dn-091.site-000.jotta.no">
  <path xml:space="preserve">/havardgulldahl/Jotta/Sync</path>
  <abspath xml:space="preserve">/havardgulldahl/Jotta/Sync</abspath>
  <latestRevision>
    <number>1</number>
    <state>INCOMPLETE</state>
    <created>2014-05-22-T22:13:52Z</created>
    <modified>2014-05-19-T13:37:14Z</modified>
    <mime>video/mp4</mime>
    <mstyle>VIDEO_MP4</mstyle>
    <size>100483008</size> <!-- THIS IS THE SIZE OF WHAT'S BEEN TRANSFERED THIS FAR. havardgulldahl -->
    <md5>4d7cdab5256b72d17075ec388e467e99</md5>
    <updated>2015-05-29-T18:07:56Z</updated>
  </latestRevision>
</file>
</file>"""

    def resume(self, data):
        'Resume uploading an incomplete file, after a previous upload was interrupted. Returns new file object'
        if not hasattr(data, 'read'):
            data = StringIO(data)

        #Check that we actually know from what byte to resume.
        #If self.size === -1, it means we never got the value from the server.
        #This is perfectly normal if the file was instatiated via e.g. a file listing,
        #and not directly via JFS.getObject()
        if self.size == -1:
            log.debug('%r is an incomplete file, but .size is unknown. Refreshing the file object from server', self.path)
            self.f = self.jfs.get(self.path)

        #check if what we're asked to upload is actually the right file
        md5 = calculate_md5(data)
        if md5 != self.md5:
            raise JFSError('''MD5 hashes don't match! Are you trying to resume with the wrong file?''')
        log.debug('Resuming %s from offset %s', self.path, self.size)
        return self.jfs.up(self.path, data, resume_offset=self.size)

    @property
    def size(self):
        """Bytes uploaded of the file so far.

        Note that we only have the file size if the file was requested directly,
        not if it's part of a folder listing.
        """
        if hasattr(self.f.latestRevision, 'size'):
            return int(self.f.latestRevision.size)
        return None

class JFSFile(JFSIncompleteFile):
    'OO interface to a file, for convenient access. Type less, do more.'
    ## TODO: add <revisions> iterator for all
    """
<file name="jottacloud.sync.pdfname" uuid="37530f11-d55b-4f31-acf4-27854813cd34" time="2013-12-15-T01:11:52Z" host="dn-029.site-000.jotta.no">
  <path xml:space="preserve">/havardgulldahl/Jotta/Sync</path>
  <abspath xml:space="preserve">/havardgulldahl/Jotta/Sync</abspath>
  <currentRevision>
    <number>1</number>
    <state>COMPLETED</state>
    <created>2013-07-19-T22:59:16Z</created>
    <modified>2013-07-19-T22:59:17Z</modified>
    <mime>application/octet-stream</mime>
    <mstyle>APPLICATION_OCTET_STREAM</mstyle>
    <size>218028</size>
    <md5>e8f05ca4ebd70bc93ce2f18e26cee2a3</md5>
    <updated>2013-07-19-T22:59:31Z</updated>
  </currentRevision>
</file>

    """
    # Constants for thumb nail sizes
    BIGTHUMB='WL'
    MEDIUMTHUMB='WM'
    SMALLTHUMB='WS'
    XLTHUMB='WXL'

    def __init__(self, fileobject, jfs, parentpath): # fileobject from lxml.objectify
        self.f = fileobject
        self.jfs = jfs
        self.parentPath = parentpath

    def stream(self, chunk_size=64*1024):
        'Returns a generator to iterate over the file contents'
        #return self.jfs.stream(url='%s?mode=bin' % self.path, chunk_size=chunk_size)
        return self.jfs.stream(url=self.path, params={'mode':'bin'}, chunk_size=chunk_size)

    def read(self):
        'Get the file contents as string'
        #return self.jfs.raw('%s?mode=bin' % self.path)
        return self.jfs.raw(url=self.path, params={'mode':'bin'})
        """
            * name = 'jottacloud.sync.pdfname'
            * uuid = '37530f11-d55b-4f31-acf4-27854813cd34'
            currentRevision = None [ObjectifiedElement]
                number = 1 [IntElement]
                state = 'COMPLETED' [StringElement]
                created = '2010-11-19-T12:34:18Z' [StringElement]
                modified = '2010-11-19-T12:34:18Z' [StringElement]
                mime = 'image/jpeg' [StringElement]
                mstyle = 'IMAGE_JPEG' [StringElement]
                size = 125848 [IntElement]
                md5 = 'a0dc8233169b238681c43f9981efe8e1' [StringElement]
                updated = '2010-11-19-T12:34:28Z' [StringElement]
        """

    def readpartial(self, start, end):
        'Get a part of the file, from start byte to end byte (integers)'
        #return self.jfs.raw('%s?mode=bin' % self.path,
        return self.jfs.raw(url=self.path, params={'mode':'bin'},
                            # note that we deduct 1 from end because
                            # in http Range requests, the end value is included in the slice,
                            # whereas in python, it is not
                            extra_headers={'Range':'bytes=%s-%s' % (start, end-1)})

    def write(self, data):
        'Put, possibly replace, file contents with (new) data'
        if not hasattr(data, 'read'):
            data = StringIO(data)
        self.jfs.up(self.path, data)

    def share(self):
        'Enable public access at secret, share only uri, and return that uri'
        # This is what jottacloud.com does
        # HTTP GET to
        #
        # https://www.jottacloud.com/web/share/{sync,backup,archive}/list/[folder uuid]/@[base64-encodded basename]?t=[timestamp]
        #
        # e.g.
        # https://www.jottacloud.com/web/share/backup/list/002c7707c2b27604dc4670660961a33a648/@YmzDpWLDpnIudXRmOC50eHQ=?t=1465934084756
        # https://www.jottacloud.com/web/share/backup/list/002c7707c2b27604dc4670660961a33a648/@YmzDpWLDpnIudXRmOC50eHQ=?t=1465934084756
        #
        #
        raise NotImplementedError('Jottacloud has changed the sharing API. Please use jottacloud.com in a browser, for now.')
        url = 'https://www.jottacloud.com/rest/webrest/%s/action/enableSharing' % self.jfs.username
        data = {'paths[]':self.path.replace(JFS_ROOT, ''),
                'web':'true',
                'ts':int(time.time()),
                'authToken':0}
        r = self.jfs.post(url, content=data)
        return r

    def restore(self):
        'Restore the file'
        #
        #
        # As of 2016-06-15, Jottacloud.com has changed their restore api
        # To restore, this is what's done
        #
        # HTTP POST to https://www.jottacloud.com/web/restore/trash/list
        # Data:
        #     hash:undefined
        #     files:@0025d37be5329a18eece18dd93f793509e8_dGVzdF9kZWxldGUudHh0
        #
        # where `files` is a comma separated list, and each item is constructed thus:
        #  @<uuid of path>_<base64 encoded file name>
        #
        if not self.deleted:
            raise JFSError('Tried to restore a not deleted file')
        raise NotImplementedError('Jottacloud has changed the restore API. Please use jottacloud.com in a browser, for now.') #  TODO: figure out how to solve this
        url = 'https://www.jottacloud.com/rest/webrest/%s/action/restore' % self.jfs.username
        data = {'paths[]': self.path.replace(JFS_ROOT, ''),
                'web': 'true',
                'ts': int(time.time()),
                'authToken': 0}
        r = self.jfs.post(url, content=data)
        return r

    def hard_delete(self):
        'Deletes without possibility to restore'
        url = 'https://www.jottacloud.com/rest/webrest/%s/action/delete' % self.jfs.username
        data = {'paths[]': self.path.replace(JFS_ROOT, ''),
                'web': 'true',
                'ts': int(time.time()),
                'authToken': 0}
        r = self.jfs.post(url, content=data)
        return r

    def delete(self):
        'Delete this file and return the new, deleted JFSFile'
        #url = '%s?dl=true' % self.path
        r = self.jfs.post(url=self.path, params={'dl':'true'})
        return r

    def rename(self, newpath):
        "Move file to a new name, possibly a whole new path"
        # POST https://www.jottacloud.com/jfs/**USERNAME**/Jotta/Sync/testFolder/testFile.txt?mv=/**USERNAME**/Jotta/Sync/testFolder/renamedTestFile.txt
        #url = '%s?mv=/%s%s' % (self.path, self.jfs.username, newpath)
        r = self.jfs.post(url=self.path,
                          params={'mv':'/%s%s' % (self.jfs.username, newpath)},
                          extra_headers={'Content-Type':'application/octet-stream'})
        return r

    def thumb(self, size=BIGTHUMB):
        '''Get a thumbnail as string or None if the file isnt an image

        size would be one of JFSFile.BIGTHUMB, .MEDIUMTHUMB, .SMALLTHUMB or .XLTHUMB'''
        if not self.is_image():
            return None
        if not size in (self.BIGTHUMB, self.MEDIUMTHUMB, self.SMALLTHUMB, self.XLTHUMB):
            raise JFSError('Invalid thumbnail size: %s for image %s' % (size, self.path))
        #return self.jfs.raw('%s?mode=thumb&ts=%s' % (self.path, size))
        return self.jfs.raw(url=self.path,
                            params={'mode':'thumb', 'ts':size})

    @property
    def revisionNumber(self):
        'return int of current revision'
        return int(self.f.currentRevision.number)

    @property
    def created(self):
        'return datetime.datetime'
        return dateutil.parser.parse(str(self.f.currentRevision.created))

    @property
    def modified(self):
        'return datetime.datetime'
        return dateutil.parser.parse(str(self.f.currentRevision.modified))

    @property
    def updated(self):
        'return datetime.datetime'
        return dateutil.parser.parse(str(self.f.currentRevision.updated))

    @property
    def size(self):
        'return int of size in bytes'
        return int(self.f.currentRevision.size)

    @property
    def md5(self):
        return str(self.f.currentRevision.md5)

    @property
    def mime(self):
        return unicode(self.f.currentRevision.mime)

    @property
    def state(self):
        return unicode(self.f.currentRevision.state)

class JFSMountPoint(JFSFolder):
    'OO interface to a mountpoint, for convenient access. Type less, do more.'
    def __init__(self, mountpointobject, jfs, parentpath): # folderobject from lxml.objectify
        super(JFSMountPoint, self).__init__(mountpointobject, jfs, parentpath)
        self.folder = mountpointobject # name it 'folder' because of inheritance

    def delete(self):
        "override inherited method that makes no sense here"
        raise JFSError('Cant delete a mountpoint')

    def rename(self, newpath):
        "override inherited method that makes no sense here"
        raise JFSError('Cant rename a mountpoint')

    @property
    def name(self):
        return unicode(self.folder.name)

    @property
    def size(self):
        'Return int of size in bytes'
        return int(self.folder.size)

    @property
    def modified(self):
        'Return datetime.datetime'
        return dateutil.parser.parse(str(self.folder.modified))

class JFSDevice(object):
    '''OO interface to a device, for convenient access. Type less, do more.


    Note that sometimes we cheat a little and instantiate this object with only the elements
    available from <user>, in which case some elements aren't there.'''
    """ raw xml example:
    <device time="2014-02-20-T21:02:42Z" host="dn-036.site-000.jotta.no">
  <name xml:space="preserve">laptop</name>
  <type>LAPTOP</type>
  <sid>d831efc4-f885-4d97-bd8d-</sid>
  <size>371951820971</size>
  <modified>2014-02-20-T14:03:52Z</modified>
  <!-- the following elements are only available if we get the metadata from
       the http path explicitly.
       you won't find it here under the <user/> element -->
  <user>hgl</user>
  <mountPoints>
    <mountPoint>
      <name xml:space="preserve">backup</name>
      <size>372544055053</size>
      <modified>2014-02-20-T14:03:52Z</modified>
    </mountPoint>
    <mountPoint>
      <name xml:space="preserve">Desktop</name>
      <size>581758</size>
      <modified>2010-11-12-T20:44:15Z</modified>
    </mountPoint>
    <mountPoint>
      <name xml:space="preserve">Documents</name>
      <size>417689097</size>
      <modified>2010-12-19-T22:40:16Z</modified>
    </mountPoint>
    <mountPoint>
      <name xml:space="preserve">Downloads</name>
      <size>0</size>
      <modified>2010-11-14-T21:00:07Z</modified>
    </mountPoint>
    <mountPoint>
      <name xml:space="preserve">Pictures</name>
      <size>5150529</size>
      <modified>2010-11-29-T11:13:03Z</modified>
    </mountPoint>
    <mountPoint>
      <name xml:space="preserve">Videos</name>
      <size>13679997</size>
      <modified>2010-11-29-T11:13:46Z</modified>
    </mountPoint>
  </mountPoints>
  <metadata first="" max="" total="6" num_mountpoints="6"/>
</device>
"""
    def __init__(self, deviceobject, jfs, parentpath): # deviceobject from lxml.objectify
        self.dev = deviceobject
        self._jfs = jfs
        self.parentPath = parentpath
        self.mountPoints = {unicode(mp.name):mp for mp in self.mountpointobjects()}

    def contents(self, path=None):
        """Get _all_ metadata for this device.
        Call this method if you have the lite/abbreviated device info from e.g. <user/>. """
        if isinstance(path, object) and hasattr(path, 'name'):
            log.debug("passed an object, use .'name' as path value")
            # passed an object, use .'name' as path value
            path = '/%s' % path.name
        c = self._jfs.get('%s%s' % (self.path, path or '/'))
        return c

    def mountpointobjects(self):
        try:
            return [ JFSMountPoint(obj, self._jfs, self.path) for obj in self.contents().mountPoints.iterchildren() ]
        except AttributeError:
            # there are no mountpoints. this may happen on newly created devices. see github bug#26
            return []

    def files(self, mountPoint):
        """Get an iterator of JFSFile() from the given mountPoint.

        "mountPoint" may be either an actual mountPoint element from JFSDevice.mountPoints{} or its .name. """
        if isinstance(mountPoint, six.string_types):
            # shortcut: pass a mountpoint name
            mountPoint = self.mountPoints[mountPoint]
        try:
            return [JFSFile(f, self, parentpath='%s/%s' % (self.path, mountPoint.name)) for f in self.contents(mountPoint).files.iterchildren()]
        except AttributeError as err:
            # no files at all
            return [x for x in []]

    def folders(self, mountPoint):
        """Get an iterator of JFSFolder() from the given mountPoint.

        "mountPoint" may be either an actual mountPoint element from JFSDevice.mountPoints{} or its .name. """
        if isinstance(mountPoint, six.string_types):
            # shortcut: pass a mountpoint name
            mountPoint = self.mountPoints[mountPoint]
        try:
            return [JFSFolder(f, self, parentpath='%s/%s' % (self.path, mountPoint.name)) for f in self.contents(mountPoint).folders.iterchildren()]
        except AttributeError as err:
            # no files at all
            return [x for x in []]

    def new_mountpoint(self, name):
        """Create a new mountpoint"""
        url = posixpath.join(self.path, name)
        r = self._jfs.post(url, extra_headers={'content-type': 'application/x-www-form-urlencoded'})
        return r

    @property
    def modified(self):
        'Return datetime.datetime'
        return dateutil.parser.parse(str(self.dev.modified))

    @property
    def path(self):
        return posixpath.join(self.parentPath, self.name)

    @property
    def name(self):
        return unicode(self.dev.name)

    @property
    def type(self):
        return unicode(self.dev.type)

    @property
    def size(self):
        'Return int of size in bytes'
        return int(self.dev.size)

    @property
    def sid(self):
        return unicode(self.dev.sid)

class JFSenableSharing(object):
    'wrap enableSharing element in a python class'
    """<enableSharing>
  <files>
    <file name="V1B.docx" uuid="d4490ff3-505c-4ecd-9994-583a6668d3b9">
      <publicURI>33cb006a8ec6493a9dabab48503d022b</publicURI>
      <currentRevision>
        <number>1</number>
        <state>COMPLETED</state>
        <created>2014-10-08-T17:26:12Z</created>
        <modified>2014-10-08-T17:26:12Z</modified>
        <mime>application/msword</mime>
        <mstyle>APPLICATION_MSWORD</mstyle>
        <size>12882</size>
        <md5>5074ad00d3d97f9b938c46c78a97e817</md5>
        <updated>2014-10-08-T15:27:10Z</updated>
      </currentRevision>
    </file>
  </files>
</enableSharing>"""
    def __init__(self, sharing, jfs): # deviceobject from lxml.objectify
        self.sharing = sharing
        self.jfs = jfs

    def sharedFiles(self):
        'iterate over shared files and get their public URI'
        for f in self.sharing.files.iterchildren():
            yield (f.attrib['name'], f.attrib['uuid'],
                'https://www.jottacloud.com/p/%s/%s' % (self.jfs.username, f.publicURI.text))

class JFSsearchresult(object):
    'wrap searchresult element in a python class'
    """<searchresult  time="2016-06-14-T22:53:43Z" host="dn-098">
  <files>
    <file name="testfile_up_and_readpartial.txt" uuid="f2d204a3-c0fd-4138-86c7-76b12f03a5a6">
    <path xml:space="preserve">/**/Jotta/Archive</path>
    <abspath xml:space="preserve">/**/Jotta/Archive</abspath>
    <currentRevision>
    <number>461</number>
    <state>COMPLETED</state>
    <created>2016-06-14-T22:53:26Z</created>
    <modified>2016-06-14-T22:53:26Z</modified>
    <mime>text/plain</mime>
    <mstyle>text/plain</mstyle>
    <size>347</size>
    <md5>0c963adda33466d565d6f3395490eaee</md5>
    <updated>2016-06-14-T22:53:26Z</updated>
    </currentRevision>
    </file>
  </files>
</enableSharing>"""
    def __init__(self, searchresult, jfs): # deviceobject from lxml.objectify
        self.searchresult = searchresult
        self.jfs = jfs

    @property
    def size(self):
        'Return datetime of search time stamp'
        return dateutil.parser.parse(str(self.searchresult.attrib['time']))

    def files(self):
        'iterate over found files'
        for _f in self.searchresult.files.iterchildren():
            yield ProtoFile.factory(_f, jfs=self.jfs, parentpath=unicode(_f.abspath))



class JFS(object):
    def __init__(self, auth=None):
        from requests.auth import HTTPBasicAuth
        self.apiversion = '2.2' # hard coded per october 2014
        self.session = requests.Session() # create a session for connection pooling, ssl keepalives and cookie jar
        self.session.stream = True
        if not auth:
            auth = get_auth_info()
        self.username, password = auth
        self.session.auth = HTTPBasicAuth(self.username, password)
        self.session.verify = certifi.where()
        self.session.headers =  {'User-Agent':'jottalib %s (https://github.com/havardgulldahl/jottalib)' % (__version__, ),
                                 'X-JottaAPIVersion': self.apiversion,
                                }
        self.rootpath = JFS_ROOT + self.username
        self.fs = self.get(self.rootpath)

    def close(self):
        self.session.close()

    def escapeUrl(self, url):
        if isinstance(url, unicode):
            url = url.encode('utf-8') # urls have to be bytestrings
        return quote(url, safe=self.rootpath)
        # TODO: remove rest of this hack after tests pass
        separators = [
            '?dl=true',
            '?mkDir=true',
            '?dlDir=true',
            '?mvDir=',
            '?mv=',
            '?mode=list',
            '?mode=bin',
            '?mode=thumb&ts='
        ]
        # TODO: replace this buggy thing with proper param support, that requests will encode for us
        # for free
        separator = separators[0]
        for sep in separators:
            if sep in url:
                separator = sep
                break

        urlparts = url.rsplit(separator, 1)
        if(len(urlparts) == 2):
            url = quote(urlparts[0], safe=self.rootpath) + separator + urlparts[1]
        else:
            url = quote(urlparts[0], safe=self.rootpath)
        return url

    def request(self, url, extra_headers=None, params=None):
        'Make a GET request for url, with or without caching'
        if not url.startswith('http'):
            # relative url
            url = self.rootpath + url
        log.debug("getting url: %r, extra_headers=%r, params=%r", url, extra_headers, params)
        if extra_headers is None: extra_headers={}
        r = self.session.get(url, headers=extra_headers, params=params)

        if r.status_code in ( 500, ):
            raise JFSError(r.reason)
        return r

    def raw(self, url, extra_headers=None, params=None):
        'Make a GET request for url and return whatever content we get'
        r = self.request(url, extra_headers=extra_headers, params=params)
        # uncomment to dump raw xml
#         with open('/tmp/%s.xml' % time.time(), 'wb') as f:
#             f.write(r.content)

        if not r.ok:
            o = lxml.objectify.fromstring(r.content)
            JFSError.raiseError(o, url)
        return r.content

    def get(self, url, params=None):
        'Make a GET request for url and return the response content as a generic lxml.objectify object'
        url = self.escapeUrl(url)

        content = StringIO(self.raw(url, params=params))
        # We need to make sure that the xml fits in available memory before we parse
        # with lxml.objectify.fromstring(), or else it will bomb out.
        # If it is too big, we need to buffer it to disk before we run it through objectify. see #87
        #
        # get length of buffer
        content.seek(0,2)
        contentlen = content.tell()
        content.seek(0)

        MAX_BUFFER_SIZE=1024*1024*200 # 200MB. TODO: find a way to compute this
        if contentlen > MAX_BUFFER_SIZE:
            # xml is too big to parse with lxml.objectify.fromstring()
            contentfile = tempfile.NamedTemporaryFile()
            contentfile.write(content.read())
            o = lxml.objectify.parse(contentfile)
        else:
            o = lxml.objectify.fromstring(content.getvalue())
        if o.tag == 'error':
            JFSError.raiseError(o, url)
        return o

    def getObject(self, url_or_requests_response, params=None):
        'Take a url or some xml response from JottaCloud and wrap it up with the corresponding JFS* class'
        if isinstance(url_or_requests_response, requests.models.Response):
            # this is a raw xml response that we need to parse
            url = url_or_requests_response.url
            o = lxml.objectify.fromstring(url_or_requests_response.content)
        else:
            # this is an url that we need to fetch
            url = url_or_requests_response
            o = self.get(url, params=params) # (.get() will parse this for us)

        parent = os.path.dirname(url).replace('up.jottacloud.com', 'www.jottacloud.com')
        if o.tag == 'error':
            JFSError.raiseError(o, url)
        elif o.tag == 'device': return JFSDevice(o, jfs=self, parentpath=parent)
        elif o.tag == 'folder': return JFSFolder(o, jfs=self, parentpath=parent)
        elif o.tag == 'mountPoint': return JFSMountPoint(o, jfs=self, parentpath=parent)
        elif o.tag == 'restoredFiles': return JFSFile(o, jfs=self, parentpath=parent)
        elif o.tag == 'deleteFiles': return JFSFile(o, jfs=self, parentpath=parent)
        elif o.tag == 'file':
            return ProtoFile.factory(o, jfs=self, parentpath=parent)
#             try:
#                 if o.latestRevision.state == 'INCOMPLETE':
#                     return JFSIncompleteFile(o, jfs=self, parentpath=parent)
#                 elif o.latestRevision.state == 'CORRUPT':
#                     return JFSCorruptFile(o, jfs=self, parentpath=parent)
#             except AttributeError:
#                 return JFSFile(o, jfs=self, parentpath=parent)
        elif o.tag == 'enableSharing': return JFSenableSharing(o, jfs=self)
        elif o.tag == 'user':
            self.fs = o
            return self.fs
        elif o.tag == 'filedirlist': return JFSFileDirList(o, jfs=self, parentpath=parent)
        elif o.tag == 'searchresult': return JFSsearchresult(o, jfs=self)
        raise JFSError("invalid object: %s <- %s" % (repr(o), url_or_requests_response))

    def getLatest(self, files=10, sort=None):
        'Yield a list of the n latest files (the server minimum default is 10), optionally sorted by `sort`.'
        url = '/Jotta/Latest'
        params = {'sort': 'updated', 'max':files, 'web':'true'}
        result = self.getObject(self.request(url, params=params))
        for _f in result.files():
            yield _f


    def stream(self, url, params=None, chunk_size=64*1024):
        'Iterator to get remote content by chunk_size (bytes)'
        r = self.request(url, params=params)
        for chunk in r.iter_content(chunk_size):
            yield chunk

    def post(self, url, content='', files=None, params=None, extra_headers={}, upload_callback=None):
        'HTTP Post files[] or content (unicode string) to url'
        if not url.startswith('http'):
            # relative url
            url = self.rootpath + url

        log.debug('posting content (len %s) to url %s', len(content) if content is not None else '?', url)
        headers = self.session.headers.copy()
        headers.update(**extra_headers)

        if not files is None:
            m = requests_toolbelt.MultipartEncoder(fields=files)
            if upload_callback is not None:
                m_len = m.len # compute value for callback closure
                def callback(monitor):
                    upload_callback(monitor, m_len)

                m = requests_toolbelt.MultipartEncoderMonitor(m, callback)
            headers['content-type'] = m.content_type
        else:
            m = content
        url = self.escapeUrl(url)
        r = self.session.post(url, data=m, params=params, headers=headers)
        if not r.ok:
            log.warning('HTTP POST failed: %s', r.text)
            raise JFSError(r.reason)
        return self.getObject(r) # return a JFS* class

    def up(self, path, fileobject, upload_callback=None, resume_offset=None):
        "Upload a fileobject to path, HTTP POST-ing to up.jottacloud.com, using the JottaCloud API"
        """

        *** WHAT DID I DO?: created file
        ***

        POST https://up.jottacloud.com/jfs/**USERNAME**/Jotta/Sync/testFolder/testFile.txt?cphash=d41d8cd98f00b204e9800998ecf8427e HTTP/1.1
        User-Agent: Desktop_Jottacloud 3.0.22.203 Windows_8 6.2.9200 x86_64
        Authorization: Basic ******************
        X-JottaAPIVersion: 2.2
        X-Jfs-DeviceName: **CENSORED**
        JCreated: 2014-10-26T12:33:09Z+00:00
        JModified: 2014-10-26T12:33:09Z+00:00
        JMd5: d41d8cd98f00b204e9800998ecf8427e
        JSize: 0
        jx_csid: dOq1NCRer6uxuR/bFxihasj4QzBU3Tn7S2jVF1CE71YW1fGhxPFYYsw2T0XYjnJBtxKQzhWixmg+u5kp8bJtvMpIFHbhSDmPPSk+PVBf2UdFhXxli4YEII9a97eO4XBfn5QWAV1LJ2Z9l59jmnLkJQgfOyexkuQbxHdSLgQPXu8=
        jx_lisence: M1v3p31oQf2OXvyAn2GvfS2I2oiMXrw+cofuMVHHI/2K+wlxhj22VkON6fN6fJMsGNcMzvcFYfmKPgL0Yf8TCO5A/6ULk6N8LctY3+fPegx+Jgbyc4hh0IXwnOdqa+UZ6Lg1ub4VXr5XnX3P3IxeVDg0VbcJnzv4TbFA+oMXmfM=
        Content-Type: application/octet-stream
        Content-Length: 0
        Connection: Keep-Alive
        Accept-Encoding: gzip
        Accept-Language: nb-NO,en,*
        Host: up.jottacloud.com
        """
        url = path.replace('www.jottacloud.com', 'up.jottacloud.com')
        # Calculate file length
        fileobject.seek(0,2)
        contentlen = fileobject.tell()

        # Rewind read head to correct offset
        # If we're resuming an incomplete upload, continue from that offset
        try:
            fileobject.seek(resume_offset)
        except TypeError as e:
            if resume_offset is None:
                fileobject.seek(0)
        except IOError as e:
            log.exception(e)
            log.warning('Could not seek to file offset %r, re-starting upload of %r from 0',
                        resume_offset,
                        url)
            fileobject.seek(0)


        # Calculate file md5 hash
        md5hash = calculate_md5(fileobject)

        log.debug('posting content (len %s, hash %s) to url %r', contentlen, md5hash, url)
        try:
            mtime = os.path.getmtime(fileobject.name)
            timestamp = datetime.datetime.fromtimestamp(mtime).isoformat()
        except Exception as e:
            log.exception('Problems getting mtime from fileobjet: %r', e)
            timestamp = datetime.datetime.now().isoformat()
        params = {'cphash': md5hash}
        m = requests_toolbelt.MultipartEncoder({
             'md5': ('', md5hash),
             'modified': ('', timestamp),
             'created': ('', timestamp),
             'file': (os.path.basename(url), fileobject, 'application/octet-stream'),
        })
        headers = {'JMd5':md5hash,
                   'JCreated': timestamp,
                   'JModified': timestamp,
                   'X-Jfs-DeviceName': 'Jotta',
                   'JSize': str(contentlen), # headers have to be strings or bytes , cf #122
                   'jx_csid': '',
                   'jx_lisence': '',
                   'content-type': m.content_type,
                   }
        fileobject.seek(0) # rewind read index for requests.post
        files = {'md5': ('', md5hash),
                 'modified': ('', timestamp),
                 'created': ('', timestamp),
                 'file': (os.path.basename(url), fileobject, 'application/octet-stream')}
        return self.post(url, None, files=files, params=params, extra_headers=headers, upload_callback=upload_callback)

    def new_device(self, name, type):
        """Create a new (backup) device on jottacloud. Types can be one of
        ['workstation', 'imac', 'laptop', 'macbook', 'ipad', 'android', 'iphone', 'windows_phone']
        """
        # at least android client also includes a "cid" with is derived from the unique device id
        # and encrypted with a public key in the apk.  The field appears to be optional
        url = posixpath.join(self.rootpath, name)
        r = self.post(url, {'type': type})
        return r

    # property overloading
    @property
    def devices(self):
        'return generator of configured devices'
        return self.fs is not None and [JFSDevice(d, self, parentpath=self.rootpath) for d in self.fs.devices.iterchildren()] or [x for x in []]

    @property
    def locked(self):
        'return bool'
        return bool(self.fs.locked) if self.fs is not None else None

    @property
    def read_locked(self):
        'return bool'
        return bool(self.fs['read-locked']) if self.fs is not None else None

    @property
    def write_locked(self):
        'return bool'
        return bool(self.fs['write-locked']) if self.fs is not None else None

    @property
    def capacity(self):
        'Return int of storage capacity in bytes. A value of -1 means "unlimited"'
        return int(self.fs.capacity) if self.fs is not None else 0

    @property
    def usage(self):
        'Return int of storage usage in bytes'
        return int(self.fs.usage) if self.fs is not None else 0




#!/usr/bin/env python
# encoding: utf-8
"""A service to act on changes on local tree (using inotify or OS equivalent) and update JottaCloud tree accordingly.

"""
#
# This file is part of jottalib.
#
# jottalib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jottalib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jottalib.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015 Håvard Gulldahl <havard@gulldahl.no>

import time, os, os.path, sys, logging, argparse, posixpath

log = logging.getLogger(__name__)

from watchdog.observers import Observer # pip install watchdog
from watchdog.utils import platform
from watchdog.events import LoggingEventHandler, FileSystemEventHandler

from clint.textui import progress, puts, colored

from jottalib.JFS import JFS
from jottalib import jottacloud, __version__
from jottalib.contrib.readlnk import readlnk




"""


1. monitor a directory
2. if a new file enters, upload it to the cloud
2b. if --archive, delete it from the directory
3. if a file is renamed, rename it in the cloud
4. if a file is changed, upload the new file
5. if a file is deleted...

ARCHIVE

SHARE

SYNC
keep a tree in sync


"""

class ArchiveEventHandler(FileSystemEventHandler):
    '''Handles Archive events. Heuristics for this handler:

    new file: upload to cloud and delete locally

    Pro tip: If your create a symlink or a .lnk reference to the file you want to upload,
             the upload will start straight away. Works great for big files.

    '''
    mode = 'Archive'

    def __init__(self, jfs, topdir, jottaroot=None):
        super(ArchiveEventHandler, self).__init__()
        self.jfs = jfs
        self.topdir = topdir
        self.jottaroot = jottaroot and jottaroot or ('/Jotta/%s' % self.mode)

    def get_jottapath(self, p, filename=None):
        rel = os.path.relpath(p, self.topdir) # strip leading path
        parts = [self.jottaroot, ] + list(os.path.split(rel)) # explode path to normalize OS path separators
        if filename is not None:
            parts.pop()  # remove original filename
            parts.append(filename)  # and replace with provided one
        return posixpath.join(*parts)  # return url

    def on_modified(self, event, dry_run=False, remove_uploaded=True):
        'Called when a file (or directory) is modified. '
        super(ArchiveEventHandler, self).on_modified(event)
        src_path = event.src_path
        if event.is_directory:
            if not platform.is_darwin():
                log.info("event is a directory, safe to ignore")
                return
            # OSX is behaving erratically and we need to paper over it.
            # the OS only reports every other file event,
            # but always fires off a directory event when a file has changed. (OSX 10.9 tested)
            # so we need to find the actual file changed and then go on from there
            files = [event.src_path+"/"+f for f in os.listdir(event.src_path)]
            try:
                src_path = max(files, key=os.path.getmtime)
            except (OSError, ValueError) as e: # broken symlink or directory empty
                return
        log.info('Modified file detectd: %s', src_path)
        #
        # we're receiving events at least two times: on file open and on file close.
        # OSes might report even more
        # we're only interested in files that are closed (finished), so we try to
        # open it. if it is locked, we can infer that someone is still writing to it.
        # this works on platforms: windows, ... ?
        # TODO: investigate http://stackoverflow.com/a/3876461 for POSIX support
        try:
            open(src_path)   # win exclusively
            os.open(src_path, os.O_EXLOCK) # osx exclusively
        except IOError: # file is not finished
            log.info('File is not finished')
            return
        except AttributeError: # no suuport for O_EXLOCK (only BSD)
            pass
        return self._new(src_path, dry_run, remove_uploaded)

    def on_created(self, event, dry_run=False, remove_uploaded=True):
        'Called when a file (or directory) is created. '
        super(ArchiveEventHandler, self).on_created(event)
        log.info("created: %s", event)

    def _new(self, src_path, dry_run=False, remove_uploaded=False):
            'Code to upload'
            # are we getting a symbolic link?
            if os.path.islink(src_path):
                sourcefile = os.path.normpath(os.path.join(self.topdir, os.readlink(src_path)))
                if not os.path.exists(sourcefile): # broken symlink
                    log.error("broken symlink %s->%s", src_path, sourcefile)
                    raise IOError("broken symliknk %s->%s", src_path, sourcefile)
                jottapath = self.get_jottapath(src_path, filename=os.path.basename(sourcefile))
            elif os.path.splitext(src_path)[1].lower() == '.lnk':
                # windows .lnk
                sourcefile = os.path.normpath(readlnk(src_path))
                if not os.path.exists(sourcefile): # broken symlink
                    log.error("broken fat32lnk %s->%s", src_path, sourcefile)
                    raise IOError("broken fat32lnk %s->%s", src_path, sourcefile)
                jottapath = self.get_jottapath(src_path, filename=os.path.basename(sourcefile))
            else:
                sourcefile = src_path
                if not os.path.exists(sourcefile): # file not exis
                    log.error("file does not exist: %s", sourcefile)
                    raise IOError("file does not exist: %s", sourcefile)
                jottapath = self.get_jottapath(src_path)

            log.info('Uploading file %s to %s', sourcefile, jottapath)
            if not dry_run:
                if not jottacloud.new(sourcefile, jottapath, self.jfs):
                    log.error('Uploading file %s failed', sourcefile)
                    raise
            if remove_uploaded:
                log.info('Removing file after upload: %s', src_path)
                if not dry_run:
                    os.remove(src_path)

class ShareEventHandler(FileSystemEventHandler):
    '''Handles Share events. Heuristics for this handler:

    new file: upload to cloud and replace contents with public share url
    delete file: delete from cloud
    '''
    def __init__(self, jfs, topdir, jottaroot=None):
        raise NotImplementedError

class SyncEventHandler(FileSystemEventHandler):
    '''Handles Sync events. Heuristics for this handler:

    new file: upload it to the cloud
    file is renamed: rename it in the cloud
    file is changed: upload the new file
    file is deleted: delete it from the cloud
    '''
    def __init__(self, jfs, topdir, jottaroot=None):
        raise NotImplementedError

def humanizeFileSize(size):
    size = abs(size)
    if (size==0):
        return "0B"
    units = ['B','KiB','MiB','GiB','TiB','PiB','EiB','ZiB','YiB']
    p = math.floor(math.log(size, 2)/10)
    return "%.3f%s" % (size/math.pow(1024,p),units[int(p)])


def filemonitor(topdir, mode, jfs):
    errors = {}
    def saferun(cmd, *args):
        log.debug('running %s with args %s', cmd, args)
        try:
            return apply(cmd, args)
        except Exception as e:
            puts(colored.red('Ouch. Something\'s wrong with "%s":' % args[0]))
            log.exception('SAFERUN: Got exception when processing %s', args)
            errors.update( {args[0]:e} )
            return False

    if mode == 'archive':
        event_handler = ArchiveEventHandler(jfs, topdir)
    elif mode == 'sync':
        event_handler = SyncEventHandler(jfs, topdir)
        #event_handler = LoggingEventHandler()
    elif mode == 'share':
        event_handler = ShareEventHandler(jfs, topdir)
    observer = Observer()
    observer.schedule(event_handler, topdir, recursive=True)
    observer.start()
    try:
        puts(colored.green('Starting JottaCloud monitor'))
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
        puts(colored.red('JottaCloud monitor stopped'))
    observer.join()



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of jottafs.
#
# jottafs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# jottafs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with jottafs.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014,2015 Håvard Gulldahl <havard@gulldahl.no>

from setuptools import setup

import os
import sys
sys.path.insert(0, './src')

from jottalib import __version__

try:
    with open('README.txt') as f:
        long_desc = f.read()
except:
    long_desc = ''

setup(name='jottalib',
      version=__version__,
      license='GPLv3',
      description='A library and tools to access the JottaCloud API',
      long_description=long_desc,
      author=u'Håvard Gulldahl',
      author_email='havard@gulldahl.no',
      url='https://github.com/havardgulldahl/jottalib',
      package_dir={'':'src'},
      packages=['jottalib', 'jottalib.contrib'],
      install_requires=['requests',
                        'requests_toolbelt',
                        'certifi',
                        'clint',
                        'python-dateutil',
                        'humanize',
                        'lxml',
                        'six',
                        'chardet',
      ],
      # see https://pythonhosted.org/setuptools/setuptools.html#declaring-extras-optional-features-with-their-own-dependencies
      extras_require={
          'Qt':  ['python-qt4',],
          'FUSE':  ['fusepy',],      # required for jotta-fuse
          'monitor': ['watchdog',],  # required for jotta-monitor
          'scanner': ['xattr',],     # optional for jotta-scanner
      },
      entry_points={
          'console_scripts': [
              'jotta-download = jottalib.cli:download',
              'jotta-fuse = jottalib.cli:fuse',
              'jotta-ls = jottalib.cli:ls',
              'jotta-mkdir = jottalib.cli:mkdir',
              'jotta-restore = jottalib.cli:restore',
              'jotta-rm = jottalib.cli:rm',
              'jotta-share = jottalib.cli:share',
              'jotta-upload = jottalib.cli:upload',
              'jotta-scanner = jottalib.cli:scanner',
              'jotta-monitor = jottalib.cli:monitor',
              'jotta-cat = jottalib.cli:cat',
        ]
      },
      classifiers="""Intended Audience :: Developers
Intended Audience :: End Users/Desktop
License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)
Operating System :: OS Independent
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: POSIX :: Linux
Operating System :: MacOS :: MacOS X
Programming Language :: Python :: 2.7
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: PyPy
Topic :: System :: Archiving
Topic :: System :: Archiving :: Backup
Topic :: Utilities""".split('\n'),
     )
